package io.awesome.gagtube.util.views;

import android.os.Bundle;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class ViewBinderHelper {
	
	private static final String BUNDLE_MAP_KEY = "ViewBinderHelper_Bundle_Map_Key";
	
	private Map<String, Integer> mapStates = Collections.synchronizedMap(new HashMap<>());
	private final Map<String, SwipeRevealLayout> mapLayouts = Collections.synchronizedMap(new HashMap<>());
	private final Set<String> lockedSwipeSet = Collections.synchronizedSet(new HashSet<>());
	
	private volatile boolean openOnlyOne = false;
	private final Object stateChangeLock = new Object();
	
	/**
	 * Help to save and restore open/close state of the swipeLayout. Call this method
	 * when you bind your view holder with the data object.
	 *
	 * @param swipeLayout swipeLayout of the current view.
	 * @param id          a string that uniquely defines the data object of the current view.
	 */
	public void bind(final SwipeRevealLayout swipeLayout, final String id) {
		
		if (swipeLayout.shouldRequestLayout()) {
			swipeLayout.requestLayout();
		}
		
		mapLayouts.values().remove(swipeLayout);
		mapLayouts.put(id, swipeLayout);
		
		swipeLayout.abort();
		swipeLayout.setDragStateChangeListener(state -> {
			
			mapStates.put(id, state);
			
			if (openOnlyOne) {
				closeOthers(id, swipeLayout);
			}
		});
		
		// first time binding.
		if (!mapStates.containsKey(id)) {
			mapStates.put(id, SwipeRevealLayout.STATE_CLOSE);
			swipeLayout.close(false);
		}
		
		// not the first time, then close or open depends on the current state.
		else {
			int state = mapStates.get(id);
			
			if (state == SwipeRevealLayout.STATE_CLOSE || state == SwipeRevealLayout.STATE_CLOSING || state == SwipeRevealLayout.STATE_DRAGGING) {
				swipeLayout.close(false);
			}
			else {
				swipeLayout.open(false);
			}
		}
		
		// set lock swipe
		swipeLayout.setLockDrag(lockedSwipeSet.contains(id));
	}
	
	public void saveStates(Bundle outState) {
		
		if (outState == null)
			return;
		
		Bundle statesBundle = new Bundle();
		for (Map.Entry<String, Integer> entry : mapStates.entrySet()) {
			statesBundle.putInt(entry.getKey(), entry.getValue());
		}
		
		outState.putBundle(BUNDLE_MAP_KEY, statesBundle);
	}
	
	@SuppressWarnings({"ConstantConditions"})
	public void restoreStates(Bundle inState) {
		
		if (inState == null)
			return;
		
		if (inState.containsKey(BUNDLE_MAP_KEY)) {
			HashMap<String, Integer> restoredMap = new HashMap<>();
			
			Bundle statesBundle = inState.getBundle(BUNDLE_MAP_KEY);
			Set<String> keySet = statesBundle.keySet();
			
			if (keySet != null) {
				for (String key : keySet) {
					restoredMap.put(key, statesBundle.getInt(key));
				}
			}
			
			mapStates = restoredMap;
		}
	}
	
	/**
	 * Lock swipe for some layouts.
	 *
	 * @param id a string that uniquely defines the data object.
	 */
	public void lockSwipe(String... id) {
		setLockSwipe(true, id);
	}
	
	/**
	 * Unlock swipe for some layouts.
	 *
	 * @param id a string that uniquely defines the data object.
	 */
	public void unlockSwipe(String... id) {
		setLockSwipe(false, id);
	}
	
	/**
	 * @param openOnlyOne If set to true, then only one row can be opened at a time.
	 */
	public void setOpenOnlyOne(boolean openOnlyOne) {
		this.openOnlyOne = openOnlyOne;
	}
	
	/**
	 * Open a specific layout.
	 *
	 * @param id unique id which identifies the data object which is bind to the layout.
	 */
	public void openLayout(final String id) {
		
		synchronized (stateChangeLock) {
			mapStates.put(id, SwipeRevealLayout.STATE_OPEN);
			
			if (mapLayouts.containsKey(id)) {
				final SwipeRevealLayout layout = mapLayouts.get(id);
				layout.open(true);
			}
			else if (openOnlyOne) {
				closeOthers(id, mapLayouts.get(id));
			}
		}
	}
	
	/**
	 * Close a specific layout.
	 *
	 * @param id unique id which identifies the data object which is bind to the layout.
	 */
	public void closeLayout(final String id) {
		
		synchronized (stateChangeLock) {
			mapStates.put(id, SwipeRevealLayout.STATE_CLOSE);
			
			if (mapLayouts.containsKey(id)) {
				final SwipeRevealLayout layout = mapLayouts.get(id);
				layout.close(true);
			}
		}
	}
	
	/**
	 * Close others swipe layout.
	 *
	 * @param id          layout which bind with this data object id will be excluded.
	 * @param swipeLayout will be excluded.
	 */
	private void closeOthers(String id, SwipeRevealLayout swipeLayout) {
		
		synchronized (stateChangeLock) {
			// close other rows if openOnlyOne is true.
			if (getOpenCount() > 1) {
				for (Map.Entry<String, Integer> entry : mapStates.entrySet()) {
					if (!entry.getKey().equals(id)) {
						entry.setValue(SwipeRevealLayout.STATE_CLOSE);
					}
				}
				
				for (SwipeRevealLayout layout : mapLayouts.values()) {
					if (layout != swipeLayout) {
						layout.close(true);
					}
				}
			}
		}
	}
	
	private void setLockSwipe(boolean lock, String... id) {
		
		if (id == null || id.length == 0)
			return;
		
		if (lock)
			lockedSwipeSet.addAll(Arrays.asList(id));
		else
			lockedSwipeSet.removeAll(Arrays.asList(id));
		
		for (String s : id) {
			SwipeRevealLayout layout = mapLayouts.get(s);
			if (layout != null) {
				layout.setLockDrag(lock);
			}
		}
	}
	
	private int getOpenCount() {
		
		int total = 0;
		
		for (int state : mapStates.values()) {
			if (state == SwipeRevealLayout.STATE_OPEN || state == SwipeRevealLayout.STATE_OPENING) {
				total++;
			}
		}
		
		return total;
	}
}
